/*
 * Copyright (c) 2001, MagniComp
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *      * Redistributions of source code must retain the above copyright 
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above copyright 
 *        notice, this list of conditions and the following disclaimer in 
 *        the documentation and/or other materials provided with the 
 *        distribution. 
 *      * Neither the name of the MagniComp nor the names of its 
 *        contributors may be used to endorse or promote products derived 
 *        from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * $Id: config-data.h,v 6.10 2001/10/22 20:20:07 mcooper Exp $
 * @(#)configdata.h
 */

#ifndef __configdata_h__
#define __configdata_h__

/*
 * Configuration data
 */

/*
 * Define the read and write values for the file descriptor array
 * used by pipe().
 */
#define PIPE_READ		0
#define PIPE_WRITE		1

/*
 * Directory information
 */
#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
typedef 	struct dirent		DirEntry_t;
#else
# define NAMLEN(dirent) (dirent)->d_namlen
typedef 	struct direct		DirEntry_t;
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

/*
 * FileSystem Interface type
 */
#if !defined(FSI_TYPE) && defined(HAVE_GETFSSTAT)
# define FSI_TYPE			FSI_GETFSSTAT	/* FSI_GETFSSTAT */
#endif
#if !defined(FSI_TYPE) && defined(HAVE_MNTCTL)
# define FSI_TYPE			FSI_MNTCTL	/* FSI_MNTCTL */
#endif
#if !defined(FSI_TYPE) && defined(HAVE_GETMNT)
# define FSI_TYPE			FSI_GETMNT	/* FSI_GETMNT */
#endif
#if !defined(FSI_TYPE) && defined(HAVE_GETMNTENT)
# define FSI_TYPE			FSI_GETMNTENT	/* FSI_GETMNTENT */
#endif

/*
 * Filesystem Mount Type
 */
#if !defined(FMT_TYPE)
# define FMT_TYPE			FMT_STATFS_NUM
#endif

/*
 * Setup for proper handling of wait3() status argument type
 */
#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

/*
 * Needed on systems with GNU libc (Linux)
 * because rdist looks inside the DIR opaque object
 */
#if !defined(DIRENT_ILLEGAL_ACCESS)
# define DIRENT_ILLEGAL_ACCESS 1
#endif

/*
 * Argument types
 */
#if !defined(ARG_TYPE) && defined(HAVE_STDARG_H)
# define ARG_TYPE		ARG_STDARG
#endif
#if !defined(ARG_TYPE) && defined(HAVE_VARARGS)
# define ARG_TYPE		ARG_VARARGS
#endif

/*
 * Allow us to override standard system definetions when needed
 */
typedef uid_t CHOWN_UID_T;		/* chown uid type */
typedef gid_t CHOWN_GID_T;		/* chown gid type */
typedef uid_t UID_T;			/* Must be signed */
typedef gid_t GID_T;			/* Must be signed */

/*
 * Generic pointer, used by memcpy, malloc, etc.  Usually char or void.
 */
#if defined(HAVE_VOID)
typedef void POINTER;
#else
typedef char POINTER;
#endif

/*
 * Constant type
 */
#if !defined(HAVE_CONST)
# define const				/* Define const to be nothing */
#endif

/*
 * Set process args to messages that show up when running ps(1)
 *
 * Under some OS's, the SETARGS code will cause ": is not an identifier"
 * errors for "special" commands.
 */
#define SETARGS

#endif	/* __configdata_h__ */
