/*
 * Copyright (c) 2001, MagniComp
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *      * Redistributions of source code must retain the above copyright 
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above copyright 
 *        notice, this list of conditions and the following disclaimer in 
 *        the documentation and/or other materials provided with the 
 *        distribution. 
 *      * Neither the name of the MagniComp nor the names of its 
 *        contributors may be used to endorse or promote products derived 
 *        from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * $Id: filesys.h,v 6.23 2001/10/22 20:21:36 mcooper Exp $
 * @(#)filesys.h
 */

#ifndef __filesys_h__
#define __filesys_h__

/*
 * File System information
 */

/*
 * Mount information
 */
#if	defined(HAVE_SYS_TYPES_H)
# include <sys/types.h>
#endif
#if	defined(HAVE_SYS_PARAM_H)
# include <sys/param.h>
#endif
#if	defined(HAVE_SYS_MOUNT_H)
# include <sys/mount.h>
#endif
#if	defined(HAVE_SYS_MNTCTL_H)
# include <sys/mntctl.h>
#endif
#if	defined(HAVE_MNTENT_H)
# include <mntent.h>
#endif
#if	defined(HAVE_SYS_MNTENT_H)
# include <sys/mntent.h>
#endif
#if	defined(HAVE_MNTTAB_H)
# include <mnttab.h>
#endif
#if	defined(HAVE_SYS_MNTTAB_H)
# include <sys/mnttab.h>
#endif

/*
 * Define pathname MOUNTED_FILE to be file containing list of
 * mounted filesystems.
 */
#if	!defined(MOUNTED_FILE) && defined(MOUNTED)
# define	MOUNTED_FILE		MOUNTED
#endif
#if	!defined(MOUNTED_FILE) && defined(_PATH_MOUNTED)
# define	MOUNTED_FILE		_PATH_MOUNTED
#endif
#if	!defined(MOUNTED_FILE) && defined(MNTTAB)
# define	MOUNTED_FILE		MNTTAB
#endif
#if	!defined(MOUNTED_FILE) && defined(MNT_MNTTAB)	/* HPUX */
# define 	MOUNTED_FILE		MNT_MNTTAB
#endif	/* MNT_MNTTAB */
/*
 * Default
 */
#if	!defined(MOUNTED_FILE)
# define	MOUNTED_FILE		"<none>"
#endif

/*
 * Stat Filesystem
 */
#if	defined(HAVE_SYS_STATFS_H)
# include <sys/statfs.h>
#endif
#if	defined(HAVE_SYS_STATVFS_H)
# include <sys/statvfs.h>
#endif
#if	defined(HAVE_SYS_VFS_H)
# include <sys/vfs.h>
#endif

/*
 * Define statfs type
 */
#if	!defined(STATFS_T) && defined(HAVE_STRUCT_STATFS)
# define STATFS_T			struct statfs
#endif
#if	!defined(STATFS_T) && defined(HAVE_STRUCT_STATVFS)
# define STATFS_T			struct statvfs
#endif
#if	!defined(STATFS_T) && defined(HAVE_STRUCT_FS_DATA)
# define STATFS_T			struct fs_data
# define f_bavail			fd_req.bfreen
# define f_bsize			fd_req.bsize
# define f_ffree			fd_req.gfree
#endif
#if	defined(STATFS_T)
 typedef STATFS_T			statfs_t;
#endif

#if defined(_AIX) || STATFS_TYPE == STATFS_SYSV
# define f_bavail			f_bfree
#endif

/*
 * Mount Entry definetions
 */
#ifndef METYPE_OTHER
# define METYPE_OTHER			"other"
#endif
#ifndef METYPE_NFS
# define METYPE_NFS			"nfs"
#endif
#ifndef MEFLAG_READONLY
# define MEFLAG_READONLY		0x01
#endif
#ifndef MEFLAG_IGNORE
# define MEFLAG_IGNORE			0x02
#endif

/*
 * Our internal mount entry type
 */
struct _mntent {
	char			       *me_path;	/* Mounted path */
	char			       *me_type;	/* Type of mount */
	int				me_flags;	/* Mount flags */
};
typedef struct _mntent mntent_t;

/*
 * Internal mount information type
 */
struct mntinfo {
	mntent_t			*mi_mnt;
	struct stat			*mi_statb;
	struct mntinfo			*mi_nxt;
};

/*
 * Declarations
 */
FILE	       *setmountent();
mntent_t       *getmountent();
mntent_t       *newmountent();
void		endmountent();

#endif	/* __filesys_h__ */
