/*
 * Copyright (c) 2001, MagniComp
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *      * Redistributions of source code must retain the above copyright 
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above copyright 
 *        notice, this list of conditions and the following disclaimer in 
 *        the documentation and/or other materials provided with the 
 *        distribution. 
 *      * Neither the name of the MagniComp nor the names of its 
 *        contributors may be used to endorse or promote products derived 
 *        from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef lint
static char RCSid[] = 
"$Id: protocmds.c,v 1.2 2001/10/22 20:31:02 mcooper Exp $";

static char copyright[] =
"Copyright (c) 2001 MagniComp\n";
#endif /* !lint */

/*
 * Protocol Commands
 */

#include "defs.h"
#include "protocmds.h"

ProtoCmd_t ProtoCmds[] = {
    /* 
     * Main commands 
     */
    { C_NONE,		"C_NONE",	"No OP" },
    { C_ERRMSG,		"C_ERRMSG",	"Log an error message" },
    { C_FERRMSG,	"C_FERRMSG",	"Log a fatal error message" },
    { C_NOTEMSG,	"C_NOTEMSG",	"Log a note message" },
    { C_LOGMSG,		"C_LOGMSG",	"Log a message" },
    { C_ACK,		"C_ACK",	"Acknowledge" },
    { C_SETCONFIG,	"C_SETCONFIG",	"Set configuration parameters" },
    { C_DIRTARGET,	"C_DIRTARGET",	"Set target directory name" },
    { C_TARGET,		"C_TARGET",	"Set target file name" },
    { C_RECVREG,	"C_RECVREG",	"Receive a regular file" },
    { C_RECVDIR,	"C_RECVDIR",	"Receive a directory" },
    { C_RECVSYMLINK,	"C_RECVSYMLINK","Receive a symbolic link" },
    { C_RECVHARDLINK,	"C_RECVHARDLINK","Receive a hard link" },
    { C_END,		"C_END",	"Indicate end of recieve/send" },
    { C_CLEAN,		"C_CLEAN",	"Clean up" },
    { C_QUERY,		"C_QUERY",	"Query without checking" },
    { C_SPECIAL,	"C_SPECIAL",	"Execute special command" },
    { C_CMDSPECIAL,	"C_CMDSPECIAL",	"Execute cmd special command" },
    { C_CHMOG,		"C_CHMOG",	"Chown, chgrp, chmod a file" },
    /* 
     * Command End Status 
     */
    { CE_UPDATED,	"CE_UPDATED",	"Update was performed" },
    { CE_NOUPDATE,	"CE_NOUPDATE",	"Nothing was updated" },
    /*
     * Session startup commands.
     */
    { S_VERSION,	"S_VERSION",	"Version number" },
    { S_REMOTEUSER,	"S_REMOTEUSER",	"Remote user name" },
    { S_LOCALUSER,	"S_LOCALUSER",	"Local user name" },
    { S_END,		"S_END",	"End of session startup commands" },
    /*
     * These are the commands for "set config".
     */
    { SC_FREESPACE,	"SC_FREESPACE",	"Set min free space" },
    { SC_FREEFILES,	"SC_FREEFILES",	"Set min free files" },
    { SC_HOSTNAME,	"SC_HOSTNAME",	"Set client hostname" },
    { SC_LOGGING,	"SC_LOGGING",	"Set logging options" },
    { SC_DEFOWNER,	"SC_DEFOWNER",	"Set default owner" },
    { SC_DEFGROUP,	"SC_DEFGROUP",	"Set default group" },
    /*
     * Query commands
     */
    { QC_ONNFS,		"QC_ONNFS",	"File exists & is on a NFS" },
    { QC_ONRO,		"QC_ONRO",	"File exists & is on a readonly fs" },
    { QC_NO,		"QC_NO",	"File does not exist" },
    { QC_SYM,		"QC_SYM",	"File exists & is a symlink" },
    { QC_YES,		"QC_YES",	"File does exist" },
    /*
     * Clean commands
     */
    { CC_QUERY,		"CC_QUERY",	"Query if file should be rm'ed" },
    { CC_END,		"CC_END",	"End of cleaning" },
    { CC_YES,		"CC_YES",	"File doesn't exist - remove" },
    { CC_NO,		"CC_NO",	"File does exist - don't remove" },
    /*
     * Run Command commands
     */
    { RC_FILE,		"RC_FILE",	"Name of a target file" },
    { RC_COMMAND,	"RC_COMMAND",	"Command to run" },
    /*
     * Mark end of commands
     */
    { CNULL, NULL, NULL },
};

/*
 * Return protocol command by it's numeric key
 */
extern ProtoCmd_t *ProtoCmdGetByKey(Key)
     int Key;
{
    register ProtoCmd_t *pc;

    for (pc = ProtoCmds; pc && pc->Name; ++pc)
	if (pc->Key == Key)
	    return pc;

    return (ProtoCmd_t *) NULL;
}
